# -*- coding: utf-8 -*-
"""
    Secret Sharing
    ~~~~~

    :copyright: (c) 2014 by Halfmoon Labs
    :license: MIT, see LICENSE for more details.
"""

from builtins import zip
from builtins import range
from utilitybelt import secure_randint as randint


def egcd(a, b):
    if a == 0:
        return (b, 0, 1)
    else:
        g, y, x = egcd(b % a, a)
        return (g, x - (b // a) * y, y)


def mod_inverse(k, prime):
    k = k % prime
    if k < 0:
        r = egcd(prime, -k)[2]
    else:
        r = egcd(prime, k)[2]
    return (prime + r) % prime


def random_polynomial(degree, intercept, upper_bound):
    """ Generates a random polynomial with positive coefficients.
    """
    if degree < 0:
        raise ValueError('Degree must be a non-negative number.')
    coefficients = [intercept]
    for i in range(degree):
        random_coeff = randint(0, upper_bound-1)
        coefficients.append(random_coeff)
    return coefficients


def get_polynomial_points(coefficients, num_points, prime):
    """ Calculates the first n polynomial points.
        [ (1, f(1)), (2, f(2)), ... (n, f(n)) ]
    """
    points = []
    for x in range(1, num_points+1):
        # start with x=1 and calculate the value of y
        y = coefficients[0]
        # calculate each term and add it to y, using modular math
        for i in range(1, len(coefficients)):
            exponentiation = (x**i) % prime
            term = (coefficients[i] * exponentiation) % prime
            y = (y + term) % prime
        # add the point to the list of points
        points.append((x, y))
    return points


def modular_lagrange_interpolation(x, points, prime):
    # break the points up into lists of x and y values
    x_values, y_values = list(zip(*points))
    # initialize f(x) and begin the calculation: f(x) = SUM( y_i * l_i(x) )
    f_x = 0
    for i in range(len(points)):
        # evaluate the lagrange basis polynomial l_i(x)
        numerator, denominator = 1, 1
        for j in range(len(points)):
            # don't compute a polynomial fraction if i equals j
            if i == j:
                continue
            # compute a fraction & update the existing numerator + denominator
            numerator = (numerator * (x - x_values[j])) % prime
            denominator = (denominator * (x_values[i] - x_values[j])) % prime
        # get the polynomial from the numerator + denominator mod inverse
        lagrange_polynomial = numerator * mod_inverse(denominator, prime)
        # multiply the current y & the evaluated polynomial & add it to f(x)
        f_x = (prime + f_x + (y_values[i] * lagrange_polynomial)) % prime
    return f_x



# -*- coding: utf-8 -*-
"""
    Secret Sharing
    ~~~~~

    :copyright: (c) 2014 by Halfmoon Labs
    :license: MIT, see LICENSE for more details.
"""

__version__ = '0.2.10'

from .sharing import secret_int_to_points, points_to_secret_int, \
    point_to_share_string, share_string_to_point, SecretSharer, \
    HexToHexSecretSharer, PlaintextToHexSecretSharer, \
    BitcoinToB58SecretSharer, BitcoinToB32SecretSharer, \
    BitcoinToZB32SecretSharer
from .secret_utility import SecretGenerator, SecretRecoverer



# -*- coding: utf-8 -*-
"""
    Secret Sharing
    ~~~~~

    :copyright: (c) 2014 by Halfmoon Labs
    :license: MIT, see LICENSE for more details.
"""
from __future__ import absolute_import

from builtins import zip
from builtins import object
import string

from six import integer_types
from utilitybelt import int_to_charset, charset_to_int, base58_chars, \
    base32_chars, zbase32_chars
from primes import get_large_enough_prime
from polynomials import random_polynomial, \
    get_polynomial_points, modular_lagrange_interpolation


def secret_int_to_points(secret_int, point_threshold, num_points, prime=None):
    """ Split a secret (integer) into shares (pair of integers / x,y coords).

        Sample the points of a random polynomial with the y intercept equal to
        the secret int.
    """
    if point_threshold < 2:
        raise ValueError("Threshold must be >= 2.")
    if point_threshold > num_points:
        raise ValueError("Threshold must be < the total number of points.")
    if not prime:
        prime = get_large_enough_prime([secret_int, num_points])
    if not prime:
        raise ValueError("Error! Secret is too long for share calculation!")
    coefficients = random_polynomial(point_threshold-1, secret_int, prime)
    points = get_polynomial_points(coefficients, num_points, prime)
    return points


def points_to_secret_int(points, prime=None):
    """ Join int points into a secret int.

        Get the intercept of a random polynomial defined by the given points.
    """
    if not isinstance(points, list):
        raise ValueError("Points must be in list form.")
    for point in points:
        if not isinstance(point, tuple) and len(point) == 2:
            raise ValueError("Each point must be a tuple of two values.")
        if not (isinstance(point[0], integer_types) and
                isinstance(point[1], integer_types)):
            raise ValueError("Each value in the point must be an int.")
    x_values, y_values = list(zip(*points))
    if not prime:
        prime = get_large_enough_prime(y_values)
    free_coefficient = modular_lagrange_interpolation(0, points, prime)
    secret_int = free_coefficient  # the secret int is the free coefficient
    return secret_int


def point_to_share_string(point, charset):
    """ Convert a point (a tuple of two integers) into a share string - that is,
        a representation of the point that uses the charset provided.
    """
    # point should be in the format (1, 4938573982723...)
    if '-' in charset:
        raise ValueError(
            'The character "-" cannot be in the supplied charset.')
    if not (isinstance(point, tuple) and len(point) == 2 and
            isinstance(point[0], integer_types) and
            isinstance(point[1], integer_types)):
        raise ValueError(
            'Point format is invalid. Must be a pair of integers.')
    x, y = point
    x_string = int_to_charset(x, charset)
    y_string = int_to_charset(y, charset)
    share_string = x_string + '-' + y_string
    return share_string


def share_string_to_point(share_string, charset):
    """ Convert a share string to a point (a tuple of integers).
    """
    # share should be in the format "01-d051080de7..."
    if '-' in charset:
        raise ValueError(
            'The character "-" cannot be in the supplied charset.')
    if not isinstance(share_string, str) and share_string.count('-') == 1:
        raise ValueError('Share format is invalid.')
    x_string, y_string = share_string.split('-')
    # print (share_string)
    # print (x_string)
    # print (y_string)
    if (set(x_string) - set(charset)) or (set(y_string) - set(charset)):
        raise ValueError("Share has characters that aren't in the charset.")
    
    if (isinstance(x_string, bytes)):
        temp = str(x_string)
        x_string = ""
        for i in range (2, len(temp)-1):
            x_string += temp[i]

    if (isinstance(y_string, bytes)):
        temp = str(y_string)
        y_string = ""
        for i in range (2, len(temp)-1):
            y_string += temp[i]

    x = charset_to_int(x_string, charset)
    y = charset_to_int(y_string, charset)
    return (x, y)


class SecretSharer(object):
    """ Creates a secret sharer, which can convert from a secret string to a
        list of shares and vice versa. The splitter is initialized with the
        character set of the secrets and the character set of the shares that
        it expects to be dealing with.
    """
    secret_charset = string.hexdigits[0:16]
    share_charset = string.hexdigits[0:16]

    def __init__(self):
        pass

    @classmethod
    def split_secret(cls, secret_string, share_threshold, num_shares):
        if (isinstance(secret_string, bytes)):
            temp = str(secret_string)
            secret_string = ""
            for i in range (2, len(temp)-1):
                secret_string += temp[i]
        secret_int = charset_to_int(secret_string, cls.secret_charset)
        points = secret_int_to_points(secret_int, share_threshold, num_shares)
        shares = []
        for point in points:
            shares.append(point_to_share_string(point, cls.share_charset))
        return shares

    @classmethod
    def recover_secret(cls, shares):
        points = []
        for share in shares:
            points.append(share_string_to_point(share, cls.share_charset))
        secret_int = points_to_secret_int(points)
        secret_string = int_to_charset(secret_int, cls.secret_charset)
        return secret_string


class HexToHexSecretSharer(SecretSharer):
    """ Standard sharer for converting hex secrets to hex shares.
    """
    secret_charset = string.hexdigits[0:16]
    share_charset = string.hexdigits[0:16]


class PlaintextToHexSecretSharer(SecretSharer):
    """ Good for converting secret messages into standard hex shares.
    """
    secret_charset = string.printable
    share_charset = string.hexdigits[0:16]


class BitcoinToB58SecretSharer(SecretSharer):
    """ Good for converting Bitcoin secret keys into shares that can be
        reliably printed out in any font.
    """
    secret_charset = base58_chars
    share_charset = base58_chars


class BitcoinToB32SecretSharer(SecretSharer):
    """ Good for converting Bitcoin secret keys into shares that can be
        reliably and conveniently transcribed.
    """
    secret_charset = base58_chars
    share_charset = base32_chars


class BitcoinToZB32SecretSharer(SecretSharer):
    """ Good for converting Bitcoin secret keys into shares that can be
        reliably and conveniently transcribed.
    """
    secret_charset = base58_chars
    share_charset = zbase32_chars



from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import range
from builtins import object
from sharing import SecretSharer 
import string 
import binascii 
import constant
from data_validator import is_valid 

class SecretGenerator(object): 
    def __init__(self, data, min_consensus_node = 3, total_number_of_node = 5):
        super(SecretGenerator, self).__init__()
        self.data = data
        self.min_consensus_node = min_consensus_node
        self.total_number_of_node = total_number_of_node
        self.secrets = [""] * self.total_number_of_node
        
    def is_valid_length(self, chunked_data): 
        if len(chunked_data) > constant.MAX_HEX_LENGTH :
            return False
        return True 

    def get_secrets_from_valid_hex_string(self, chunked_data): 
        if self.is_valid_length(chunked_data) == False: 
            raise ValueError("Data length must be less than " + str(constant.MAX_HEX_LENGTH) + " character") 

        if all(c in string.hexdigits for c in chunked_data.decode()) == False: 
            raise ValueError("Data must contain only hexdigits") 

        return SecretSharer.split_secret(chunked_data, self.min_consensus_node, self.total_number_of_node) 

    def get_secrets_from_hex_string(self, chunked_data): 
        data_len = len(chunked_data) 
        if data_len <= constant.MAX_HEX_LENGTH:
            return self.get_secrets_from_valid_hex_string(chunked_data)

        return "Hex string length must be less than " + str(constant.MAX_HEX_LENGTH) 

    def get_chunk(self, idx, data_len):

        chunked_data = ""
        for i in range(idx, min(data_len, idx+constant.MAX_CHUNK_LENGTH)):
            chunked_data += self.data[i]
        return chunked_data

    def add_caps(self):

        for node in range(self.total_number_of_node):
            self.secrets[node] = self.secrets[node] + "^"

    def add_chunk(self, chunkSecrets):

        for node in range(self.total_number_of_node): 
            self.secrets[node] = self.secrets[node] + chunkSecrets[node]
    
    def chunked_text_to_chunked_secret(self, idx, data_len):
        temp_data = self.get_chunk(idx, data_len)
        hex_temp_data = binascii.hexlify(temp_data.encode())
        return self.get_secrets_from_hex_string(hex_temp_data)

    def get_secrets_from_plain_text(self): 
        
        if is_valid(self.data) == False: 
            return "Data must contain only ascii character" 
        data_len = len(self.data)
        for idx in range(0, data_len, constant.MAX_CHUNK_LENGTH):

            chunkSecrets = self.chunked_text_to_chunked_secret(idx, data_len)
            if (idx > 0): self.add_caps()
            self.add_chunk(chunkSecrets)

        return self.secrets

    def run(self):
        return self.get_secrets_from_plain_text()

class SecretRecoverer(object):
    def __init__(self, secrets):
        super(SecretRecoverer, self).__init__()
        self.secrets = secrets
        self.data = None

    def recover_hex_string_secret(self, hex_string): 
        return SecretSharer.recover_secret(hex_string) 

    def special_case(self, secrets):
        temp_data = self.recover_hex_string_secret(secrets)
        recovered_data = str(binascii.unhexlify(temp_data))
        recovered_data = recovered_data[2:len(recovered_data)-1]
        return recovered_data

    def add_pieces_together(self, total_secrets):
        self.data = [[] for i in range(total_secrets)]
        for line in self.secrets:
            pieces = str(line).split('^') 
            for i in range(total_secrets):
                self.data[i].append(pieces[i])

    def decrypt_chunked_message(self, idx):
        hex_sub_key = self.recover_hex_string_secret(self.data[idx])
        sub_key = binascii.unhexlify(hex_sub_key)
        return sub_key

    def decrypt_whole_message(self, total_secrets):
        recovered_data = ""
        for i in range(total_secrets):
            temp = str(self.decrypt_chunked_message(i))
            recovered_data += temp[2:len(temp)-1]
        return str(recovered_data)

    def recover_plain_text_secret(self): 
        total_secrets = str(self.secrets[0]).count('^') + 1

        if total_secrets == 1: 
            return self.special_case(self.secrets)

        self.add_pieces_together(total_secrets)
        return self.decrypt_whole_message(total_secrets) 

    def run (self):
        return self.recover_plain_text_secret()




# -*- coding: utf-8 -*-
"""
    Secret Sharing
    ~~~~~

    :copyright: (c) 2014 by Halfmoon Labs
    :license: MIT, see LICENSE for more details.
"""


from builtins import range
def calculate_mersenne_primes():
    """ Returns all the mersenne primes with less than 500 digits.
        All primes:
        3, 7, 31, 127, 8191, 131071, 524287, 2147483647L, 2305843009213693951L,
        618970019642690137449562111L, 162259276829213363391578010288127L,
        170141183460469231731687303715884105727L,
        68647976601306097149...12574028291115057151L, (157 digits)
        53113799281676709868...70835393219031728127L, (183 digits)
        10407932194664399081...20710555703168729087L, (386 digits)
    """
    mersenne_prime_exponents = [
        2, 3, 5, 7, 13, 17, 19, 31, 61, 89, 107, 127, 521, 607, 1279
    ]
    primes = []
    for exp in mersenne_prime_exponents:
        prime = 1
        for i in range(exp):
            prime *= 2
        prime -= 1
        primes.append(prime)
    return primes

SMALLEST_257BIT_PRIME = (2**256 + 297)
SMALLEST_321BIT_PRIME = (2**320 + 27)
SMALLEST_385BIT_PRIME = (2**384 + 231)
STANDARD_PRIMES = calculate_mersenne_primes() + [
    SMALLEST_257BIT_PRIME, SMALLEST_321BIT_PRIME, SMALLEST_385BIT_PRIME
]
STANDARD_PRIMES.sort()


def get_large_enough_prime(batch):
    """ Returns a prime number that is greater all the numbers in the batch.
    """
    # build a list of primes
    primes = STANDARD_PRIMES
    # find a prime that is greater than all the numbers in the batch
    for prime in primes:
        numbers_greater_than_prime = [i for i in batch if i > prime]
        if len(numbers_greater_than_prime) == 0:
            return prime
    return None



"""
Secret Sharing
==============

"""

from setuptools import setup

setup(
    name='secretsharingng',
    version='0.2.10',
    url='https://github.com/covalent-hq/secret-sharing-ng',
    license='MIT',
    author='Halfmoon Labs',
    author_email='hello@halfmoon.io',
    description=("Tools for sharing secrets (like Bitcoin private keys), "
                 "using shamir's secret sharing scheme."),
    packages=[
        'secretsharingng',
    ],
    zip_safe=False,
    install_requires=[
        'six',
        'utilitybelt',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
    ],
)
